from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('attributes', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attributeschema',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='attributeset',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



import pytest
from model_bakery import baker

from django.contrib.contenttypes.models import ContentType
from rest_framework.test import APIClient

from huscy.attributes.models import AttributeSchema, AttributeSet
from huscy.attributes.services import create_attribute_schema


@pytest.fixture
def user(django_user_model):
    return django_user_model.objects.create_user(username='user', password='password',
                                                 first_name='Jim', last_name='Panse')


@pytest.fixture
def admin_client(admin_user):
    client = APIClient()
    client.login(username=admin_user.username, password='password')
    return client


@pytest.fixture
def client(user):
    client = APIClient()
    client.login(username=user.username, password='password')
    return client


@pytest.fixture
def anonymous_client(user):
    return APIClient()


@pytest.fixture
def schema_v2():
    return {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'}
        }
    }


@pytest.fixture
def schema_v3():
    return {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'string'}
        }
    }


@pytest.fixture
def schema_v4():
    return {
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'object'},
            'attribute2': {'type': 'string'},
            'attribute3': {'type': 'number'}
        }
    }


@pytest.fixture
def initial_attribute_schema():
    return AttributeSchema.objects.create()


@pytest.fixture
def attribute_schema_v2(initial_attribute_schema, schema_v2):
    return AttributeSchema.objects.create(schema=schema_v2)


@pytest.fixture
def attribute_schema_v3(attribute_schema_v2, schema_v3):
    return AttributeSchema.objects.create(schema=schema_v3)


@pytest.fixture
def attribute_schema_v4(attribute_schema_v3, schema_v4):
    return AttributeSchema.objects.create(schema=schema_v4)


@pytest.fixture
def subject():
    return baker.make('subjects.Subject')


@pytest.fixture
def content_type():
    return ContentType.objects.get_by_natural_key('attributes', 'attributeset')


@pytest.fixture
def pseudonym(subject, content_type):
    return baker.make(
        'pseudonyms.Pseudonym',
        subject=subject,
        content_type=content_type,
        code='1234'
    )


@pytest.fixture
def attribute_set(attribute_schema_v4, pseudonym):
    return AttributeSet.objects.create(
        pseudonym=pseudonym.code,
        attributes={
            'attribute1': {},
            'attribute2': 'any string',
            'attribute3': 4.5,
        },
        attribute_schema=attribute_schema_v4,
    )


@pytest.fixture
def attribute_set_with_categories(pseudonym):
    attribute_schema = create_attribute_schema({
        'type': 'object',
        'properties': {
            'category1': {
                'type': 'object',
                'properties': {
                    'attribute11': {'type': 'number'},
                    'attribute12': {'type': 'number'},
                }
            },
            'category2': {
                'type': 'object',
                'properties': {
                    'attribute21': {'type': 'number'},
                    'attribute22': {'type': 'number'},
                }
            }
        }
    })

    return AttributeSet.objects.create(
        pseudonym=pseudonym.code,
        attributes={
            'category1': {'attribute11': 100},
            'category2': {'attribute21': 100}
        },
        attribute_schema=attribute_schema
    )



from datetime import datetime

import pytest

from django.contrib.auth.models import Permission

from huscy.attributes.models import AttributeSchema
from huscy.attributes.services import create_attribute_schema

pytestmark = pytest.mark.django_db


@pytest.mark.freeze_time('2000-01-01T10:00:00')
def test_create_attribute_schema(django_db_reset_sequences):
    schema = dict(type='object', properties=dict())

    result = create_attribute_schema(schema)

    assert isinstance(result, AttributeSchema)
    assert result.pk == 1
    assert result.schema == schema
    assert result.created_at == datetime(2000, 1, 1, 10)


def test_create_category_permissions():
    schema = {
        'type': 'object',
        'properties': {
            'property1': {'type': 'string'},
            'property2': {'type': 'object', 'properties': {}},
            'property3': {'type': 'number'},
            'property4': {
                'type': 'object',
                'properties': {
                    'property5': {'type': 'string'},
                    'property6': {'type': 'string'},
                }
            }
        }
    }

    assert not Permission.objects.filter(codename__contains='_attribute_category_').exists()

    create_attribute_schema(schema)

    assert 4 == Permission.objects.filter(codename__contains='_attribute_category_').count()
    assert Permission.objects.filter(codename='change_attribute_category_property2').exists()
    assert Permission.objects.filter(codename='change_attribute_category_property4').exists()
    assert Permission.objects.filter(codename='view_attribute_category_property2').exists()
    assert Permission.objects.filter(codename='view_attribute_category_property4').exists()



import pytest
from model_bakery import baker

from jsonschema.exceptions import ValidationError

from huscy.attributes.services import update_attribute_set

pytestmark = pytest.mark.django_db


def test_update_attributes(attribute_set):
    attributes = attribute_set.attributes
    attributes['attribute3'] = 100.0

    update_attribute_set(attribute_set, attributes)

    attribute_set.refresh_from_db()
    assert attribute_set.attributes == {
        'attribute1': {},
        'attribute2': 'any string',
        'attribute3': 100.0,
    }


def test_update_attributes_with_invalid_data(attribute_set):
    attributes = attribute_set.attributes
    attributes['attribute3'] = 'invalid'

    with pytest.raises(ValidationError) as e:
        update_attribute_set(attribute_set, attributes)

    assert str(e.value).startswith("'invalid' is not of type 'number'")


def test_update_attributes_together_with_attribute_schema_version(attribute_set, schema_v4):
    schema_v5 = schema_v4.copy()
    schema_v5['properties']['attribute4'] = {'type': 'string', 'enum': ['a', 'b']}
    latest_attribute_schema = baker.make('attributes.AttributeSchema', schema=schema_v5)

    attributes = attribute_set.attributes
    attributes['attribute4'] = 'a'

    update_attribute_set(attribute_set, attributes, latest_attribute_schema)

    attribute_set.refresh_from_db()
    assert attribute_set.attribute_schema == latest_attribute_schema
    assert attribute_set.attributes == {
        'attribute1': {},
        'attribute2': 'any string',
        'attribute3': 4.5,
        'attribute4': 'a',
    }


def test_update_attribute_schema_to_lower_version(attribute_set, attribute_schema_v3):
    with pytest.raises(Exception) as e:
        update_attribute_set(attribute_set, attribute_set.attributes, attribute_schema_v3)

    assert str(e.value) == ('New version for attribute schema must be greater than or equals with '
                            'current attribute schema version.')


def test_partial_update_for_attributes(attribute_set_with_categories):
    attribute_set = attribute_set_with_categories

    update_attribute_set(attribute_set, {
        'category1': {'attribute12': 50},
        'category2': {'attribute22': 50}
    })

    attribute_set.refresh_from_db()
    assert attribute_set.attributes == {
        'category1': {'attribute11': 100, 'attribute12': 50},
        'category2': {'attribute21': 100, 'attribute22': 50}
    }



from django.contrib.postgres.fields import JSONField
from django.db import models


DEFAULT_ATTRIBUTE_SCHEMA = {
    'type': 'object',
    'properties': {},
}


class AttributeSchema(models.Model):
    schema = JSONField(default=DEFAULT_ATTRIBUTE_SCHEMA)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)


def get_latest_attribute_schema_version():
    return AttributeSchema.objects.count()


class AttributeSet(models.Model):
    pseudonym = models.CharField(max_length=128, unique=True)
    attributes = JSONField(default=dict)
    attribute_schema = models.ForeignKey(AttributeSchema, on_delete=models.PROTECT,
                                         default=get_latest_attribute_schema_version)



import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import huscy.attributes.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AttributeSchema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('schema', django.contrib.postgres.fields.jsonb.JSONField(default={'properties': {}, 'type': 'object'})),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='AttributeSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pseudonym', models.CharField(max_length=128, unique=True)),
                ('attributes', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('attribute_schema', models.ForeignKey(default=huscy.attributes.models.get_latest_attribute_schema_version, on_delete=django.db.models.deletion.PROTECT, to='attributes.attributeschema')),
            ],
        ),
    ]



from rest_framework.permissions import BasePermission


class AttributeSetPermission(BasePermission):
    def has_permission(self, request, view):
        if request.method == 'GET':
            return request.user.has_perm('attributes.view_attributeset')
        elif request.method == 'PUT':
            return request.user.has_perms([
                'subjects.change_subject',
                'attributes.change_attributeset'
            ])



import pytest

from django.contrib.admin.sites import AdminSite

from huscy.attributes.admin import AttributeSetAdmin
from huscy.attributes.models import AttributeSet

pytestmark = pytest.mark.django_db


@pytest.fixture
def attribute_set_admin():
    return AttributeSetAdmin(model=AttributeSet, admin_site=AdminSite())


def test_attribute_schema(attribute_set_admin, attribute_set):
    assert attribute_set_admin._attribute_schema(attribute_set) == attribute_set.attribute_schema.id


def test_has_add_permission(attribute_set_admin):
    assert attribute_set_admin.has_add_permission(request=None) is False


def test_has_change_permission(attribute_set_admin):
    assert attribute_set_admin.has_change_permission(request=None) is False


def test_has_delete_permission(attribute_set_admin):
    assert attribute_set_admin.has_delete_permission(request=None) is False
